#ifndef DATA_SET_H
#define DATA_SET_H

#include <vector> 
#include <string> 
#include <math.h>

using namespace std;


class Vector : public vector<double> {
public:
  double distance2(const Vector& x) {
    double d = 0;
    double diff;
    for (unsigned i=0; i<size(); i++) {
      diff = (*this)[i] - x[i]; 
      d += diff * diff;
    }
    return d; 
  } 

  double distance(const Vector& x) {
    double d = 0;
    double diff;
    for (unsigned i=0; i<size(); i++) {
      diff = (*this)[i] - x[i]; 
      d += diff * diff;
    }
    return sqrt(d); 
  } 

  Vector& mean(const Vector& a,const Vector& b,double r) {
    assert( a.size() == b.size() ); 
    (*this).resize( a.size() );

    for (unsigned i=0; i<a.size(); i++) 
      (*this)[i] = (1-r)*a[i]+r*b[i]; 
    
    return *this;
  } 

  Vector& operator+=(const Vector& r) {
    for (unsigned i=0; i<size(); i++)
      (*this)[i] += r[i];
    return *this;
  } 

  Vector& operator/=(double r) { 
    for (unsigned i=0; i<size(); i++)
      (*this)[i] /= r;
    return *this;
  }
 
  void zero(int n) {
    resize(n); 
    for (unsigned i=0; i<size(); i++)
	  (*this)[i] = 0;
  } 
}; 

class TrainSample {

public:
  
  int getN() { return input.size(); }
  int getM() { return output.size(); } 
  

  Vector input;
  Vector output; 
}; 


class DataSet {
  
public:
  
  DataSet() : cursor(0), name("noname") { }; 

  void load(string& name); 
  void save(string& name); 
  void print();

  TrainSample operator[](int i) const { 
    return data[i]; 
  } 

  TrainSample& operator[](int i)  {
    return data[i]; 
  } 

  void rewind() { cursor = 0; }
  
  TrainSample& next() { return data[cursor]; cursor++; }  

  int size() { return data.size(); } 
  int getN() { return min.getN(); } 
  int getM() { return max.getM(); } 

private:
  int cursor; 
  string name;
  
  vector<TrainSample> data;
public:
  TrainSample min;  // minimum for each column 
  TrainSample max;  // maximum for each column
};  

#endif 
