#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdexcept>  
#include <iostream>   
#include <iomanip>
#include <fstream>
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>  
#endif


#include <eo>
#include <ga.h>

#include "operators.h"
#include "fitness.h"
#include "dataset.h"
#include "stat.h" 

#include "gui/gui.h" 

using namespace std;



void main_function(int argc, char **argv)
{

  // PARAMETRES
  unsigned int SEED;      // seed for random number generator
  unsigned int POP_SIZE;  // Size of population
  unsigned int MAX_GEN;  // Maximum number of generation before STOP
  double CROSS_RATE;      // Crossover rate
  double P_MUT_PER_BIT; // probability of bit-flip mutation
  double MUT_RATE;        // mutation rate
  int MUTATION_TYPE; 

  eoParser parser(argc, argv);

  eoValueParam<int> seedParam(time(0), "seed", "Random number seed", 'S');
  parser.processParam( seedParam );
  SEED = seedParam.value();

  //  eoValueParam<int> mutTypeParam(0, "mutation", "Mutation type", 'T');
  // parser.processParam( mutTypeParam );
  // MUTATION_TYPE = mutTypeParam.value();

   
  eoValueParam<unsigned int> popSizeParam(50, "popSize", "Population size",'P');
  parser.processParam( popSizeParam, "Evolution engine" );
  POP_SIZE = popSizeParam.value();

  eoValueParam<unsigned int> maxGenParam(100, "maxGen", "Maximum number of generations",'G');
  parser.processParam( maxGenParam, "Stopping criterion" );
  MAX_GEN = maxGenParam.value();

  
  eoValueParam<double> pCrossParam(0.6, "pCross", "Probability of Crossover", 'C'); 
  parser.processParam( pCrossParam, "Genetic Operators" );
  CROSS_RATE = pCrossParam.value();

  eoValueParam<double> pMutParam(0.3, "pMut", "Probability of Mutation", 'M');
  parser.processParam( pMutParam, "Genetic Operators" );
  MUT_RATE = pMutParam.value();

  double mut1_prob, mut2_prob, mut3_prob, mut4_prob, mut5_prob;
  
  eoValueParam<double> pMut1(1.0, "pMut1", "Probability of Mutation1",'n'); 
  parser.processParam( pMut1, "Mutation1"); 
  mut1_prob = pMut1.value(); 

  eoValueParam<double> pMut2(0.0, "pMut2", "Probability of Mutation2",'o'); 
  parser.processParam( pMut2, "Mutation2"); 
  mut2_prob = pMut2.value(); 

  eoValueParam<double> pMut3(0.0, "pMut3", "Probability of Mutation3",'n'); 
  parser.processParam( pMut3, "Mutation3"); 
  mut3_prob = pMut3.value(); 

  eoValueParam<double> pMut4(0.5, "pMut4", "Probability of Mutation Add",'n'); 
  parser.processParam( pMut4, "Mutation4"); 
  mut4_prob = pMut4.value(); 

  eoValueParam<double> pMut5(0.5, "pMut5", "Probability of Mutation Remove",'n'); 
  parser.processParam( pMut5, "Mutation5"); 
  mut5_prob = pMut5.value(); 

  double xover1_prob, xover2_prob; 
  
  eoValueParam<double> pXover1(1.0, "pXover1", "Probability of Xover1",'x'); 
  parser.processParam( pXover1, "Xover1"); 
  xover1_prob = pXover1.value(); 

  eoValueParam<double> pXover2(0.0, "pXover2", "Probability of Xover2",'x'); 
  parser.processParam( pXover2, "Xover2"); 
  xover2_prob = pXover2.value(); 

  
  
  eoValueParam<double> pMutPerBitParam(0.01, "pMutPerBit", "Probability of flipping 1 bit in bit-flip mutation", 'b');
  parser.processParam( pMutPerBitParam, "Genetic Operators" );
  P_MUT_PER_BIT = pMutPerBitParam.value();

  
  eoValueParam<string> saveStatsParam("fitness.log", "LogFile","A file to log statistics",'L');
  parser.processParam( saveStatsParam, "Log Statistics File" );
  string saveStatsName = saveStatsParam.value();

  eoValueParam<string> saveResultParam("result.txt", "ResultFile","A file to log result",'R');
  parser.processParam( saveResultParam, "Result File" );
  string saveResultName = saveResultParam.value();


  string str_status = parser.ProgramName() + ".status"; // default value
  eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
  parser.processParam( statusParam, "Persistence" );

  eoValueParam<string>  inputParam("25c", "TaskName", "Task.",'I');
  parser.processParam( inputParam, "Task" );
  string inputName = inputParam.value() + ".train";
  string plotName = inputParam.value() + ".plot"; 
    
  eoValueParam<int>  pK(20, "NumberOfClusters","Number of clusters",'K');
  parser.processParam( pK, "NumberOfClusters" );
  int K = pK.value();
  
  
  if (parser.userNeedsHelp()) {
    parser.printHelp(cout);
    exit(1);
  }
  if (statusParam.value() != "") {
    ofstream os(statusParam.value().c_str());
    os << parser;		// and you can use that file as parameter file
  }
    

  DataSet data; 
  data.load(inputName);

    
  //////////////////////////
  //  Random seed
  //////////////////////////
  rng.reseed(SEED);

  /////////////////////////////
  // Fitness function
  ////////////////////////////
  SumDistancesFitness eval(data);
  // SilhouetteFitness eval(data);
  

  ////////////////////////////////
  // Initilisation of population
  ////////////////////////////////
  eoPop<Keywords> pop;
  for (unsigned int igeno=0; igeno<POP_SIZE; igeno++) {
    Keywords v; 
    v.resize( rng.random( K ) );
    if (v.size()<2) 
      v.resize(2);
    for (int ivar=0; ivar<v.size(); ivar++) {
      int j = rng.random( data.size() );
      v[ivar] = data[j].input;
    }
    eval(v);           
    pop.push_back(v);  
  }
  
  // OUTPUT
  //  pop.sort();
  //  cout << "Initial Population" << endl;
  // cout << pop;
   
 
  /////////////////////////////////////
  // selection and replacement
  ///////////////////////////////////
  eoDetTournamentSelect<Keywords> selectOne; 
  // eoProportionalSelect<Keywords> selectOne;  
  // MySelect select; 
  eoSelectPerc<Keywords> select(selectOne); 
  

  // REPLACE
  double elita_percentage = 0.05; 
  eoDeterministicSaDReplacement<Keywords> replace(elita_percentage,1.0-elita_percentage);

  //  eoPlusReplacement<Keywords> replace;
  // eoSSGADetTournamentReplacement<Keywords> replace(2);

  // CROSSOVER
  KeywordsXover xover1;
  KeywordsMeanXover xover2; 

  
  eoPropCombinedQuadOp<Keywords> xover(xover1,xover1_prob);
  xover.add(xover2,xover2_prob);

  // MUTATION
  /*
  eoMonOp<Keywords>* mut = 0; 

  switch (MUTATION_TYPE) {
  case 0 : // normal mutation 
    mut = new KeywordsMut1(data.min.input, data.max.input); 
    break;
  case 1 : // kmeans
    mut = new KMeansMut(data); 
  } 
  */
  KeywordsMut1 mutation1(data.min.input, data.max.input); // neovlivnena 
  KeywordsMut2 mutation2(data.min.input, data.max.input); // ovlivnena 
  KMeansMut    mutation3(data);
  KeywordsAdd  mutation4(data); 
  KeywordsRemove mutation5; 

  eoPropCombinedMonOp<Keywords> mutation(mutation1,mut1_prob);
  mutation.add(mutation2,mut2_prob);
  mutation.add(mutation3,mut3_prob); 
  mutation.add(mutation4,mut4_prob);
  mutation.add(mutation5,mut5_prob); 

  eoDynSGATransform<Keywords> transform(xover, &CROSS_RATE, mutation, &MUT_RATE);

  
  // CHECKPOINT
  
  //  eoGenContinue<Keywords> genCont(MAX_GEN);
  //  eoFitContinue<Keywords> fitCont(2.0);

  myGui gui(&CROSS_RATE,&MUT_RATE,
	    0,100000,
	    plotName); // stopping via gui 
 
  eoCheckPoint<Keywords> checkpoint(gui);    

  eoValueParam<unsigned> generationCounter(0, "Gen.");
  eoIncrementor<unsigned> increment(generationCounter.value());
  checkpoint.add(increment);
  
  eoBestFitnessStat<Keywords> bestStat;
  eoSecondMomentStats<Keywords> SecondStat;
  eoLengthStat length;

  checkpoint.add(bestStat);
  checkpoint.add(SecondStat);
  checkpoint.add(length);

  eoFileMonitor fileMonitor(saveStatsName, " ");
  checkpoint.add(fileMonitor);
  fileMonitor.add(generationCounter);
  fileMonitor.add(bestStat);
  fileMonitor.add(SecondStat);
  fileMonitor.add(length);

  /*
  eoState outState;
  outState.registerObject(pop);
 
  eoCountedStateSaver stateSaver1(1, outState, "generation"); 
  checkpoint.add(stateSaver1);
  */

  
  /////////////////////////////////////////
  // the GA algorithm
  ////////////////////////////////////////
  eoEasyEA<Keywords> gga(checkpoint, 
		     eval, 
		     select, 
		     transform, 
		     replace);

  
  gga(pop);
  
  // OUTPUT
  pop.sort();
  cout << "FINAL Population\n" << pop[0] << endl;

  ofstream result(saveResultName.c_str());
  for (unsigned i=0; i< pop[0].size(); i++) {
    for (unsigned j=0; j< pop[0][i].size(); j++) 
      result << pop[0][i][j] << " ";
    result << endl;
  } 
  result.close();

  // delete mut;
}


int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
